module Workarea
  class Registry
    include ApplicationDocument

    field :user_id, type: String
    field :email, type: String
    field :name, type: String
    field :description, type: String
    field :privacy, type: String, default: 'public'
    field :event_date, type: Date

    embeds_many :items,
      class_name: 'Workarea::Registry::Item',
      cascade_callbacks: true,
      inverse_of: :registry

    embeds_one :address,
      class_name: 'Workarea::Address',
      as: :addressable

    validates :name, presence: true
    validates :user_id, presence: true
    validates :email, presence: true
    validates :privacy, presence: true,
      inclusion: { in: %w(public shared private) }

    scope :with_items, -> { where('items.0' => { '$exists' => true }) }
    scope :for_user, ->(user_id) { where(user_id: user_id.to_s) }
    scope :by_activity, -> { order(updated_at: :desc) }

    index({ user_id: 1 })
    index('items.product_id' => 1)
    index(name: 1)
    index(email: 1)
    index(id: 1)

    def self.search(query)
      Workarea::Registry::PublicSearch.new(
        query
      ).results
    end

    %w(public shared private).each do |privacy_level|
      define_method "#{privacy_level}?" do
        privacy == privacy_level
      end
    end

    def default_privacy?
      public? || privacy.blank?
    end

    def purchased_items
      items.select(&:purchased_by?)
    end

    def total_purchased_items
      purchased_items.pluck('quantity').sum
    end

    def total_price
      purchased_items.map{ |item| item.total_price }.sum
    end

    def unpurchased_items
      items.reject(&:purchased_by?)
    end

    def available_items
      items.reject(&:purchased_by?).select(&:available?)
    end

    def total_items
      unpurchased_items.pluck('quantity').sum
    end

    def total_items_received
      unpurchased_items.pluck('received').sum
    end

    def total_products
      unpurchased_items.size
    end

    def add_item(params)
      quantity = params.fetch(:quantity, 1).to_i
      existing_item = find_existing_item(sku: params[:sku])

      if existing_item.present?
        update_item(
          existing_item.id,
          quantity: existing_item.quantity + quantity
        )
      else
        items.build(params)
      end

      save
    end

    def add_purchased_item(params)
      existing_purchased_item = find_existing_purchased_item(params)

      if existing_purchased_item
        quantity = existing_purchased_item.quantity + params[:quantity]
        total_price = existing_purchased_item.total_price + params[:total_price]
        existing_purchased_item.update(quantity: quantity, total_price: total_price)
      else
        items.build(params)
      end

      save
    end

    def update_item(id, params)
      sku = params[:sku]
      existing_item = find_existing_item(sku: sku) if sku.present?

      if existing_item.present? && existing_item.id.to_s != id.to_s
        item = items.find(id)
        quantity = existing_item.quantity + (params[:quantity] || item.quantity)

        existing_item.update(quantity: quantity) && item.delete
      else
        items.find(id).update(params)
      end
    end

    def remove_item(id)
      item = items.detect { |item| item.id.to_s == id.to_s }
      item.nil? || item.destroy!
    end

    def mark_item_received(item)
      existing_item = find_existing_item(sku: item.sku)

      if existing_item
        received = existing_item.received + item.quantity
        existing_item.update(received: received)
      end
    end

    def find_existing_item(params = {})
      unpurchased_items.detect { |item| item.sku == params[:sku] }
    end

    def find_existing_purchased_item(params = {})
      items.detect { |item| item.sku == params[:sku] && item.purchased_by == params[:purchased_by] }
    end
  end
end
